# -*- coding: utf-8 -*-
"""
py_gql package information
"""

__title__ = "py_gql"
__description__ = "GraphQL implementation for Python"
__url__ = "https://github.com/lirsacc/py-gql"
__version__ = "0.1.0"
__author__ = "Charles Lirsac"
__author_email__ = "charles@lirsac.me"
__license__ = "MIT"
__copyright__ = "Copyright 2018 Charles Lirsac"
